// +build mage

package main

import (
	"context"
	"fmt"
	"io/ioutil"
	"log"
	"os"
	"os/exec"
	"path/filepath"
	"runtime"
	"strings"

	"github.com/jtolds/qod"
	"github.com/magefile/mage/mg"
	"github.com/pkg/errors"
)

var wd string

func init() {
	lwd, err := os.Getwd()
	qod.ANE(err)

	wd = lwd
}

const pkgBase = "git.xeserv.us/xena/route/"

func output(cmd string, args ...string) (string, error) {
	c := exec.Command(cmd, args...)
	c.Env = os.Environ()
	c.Stderr = os.Stderr
	b, err := c.Output()
	if err != nil {
		return "", errors.Wrapf(err, `failed to run %v %q`, cmd, args)
	}
	return string(b), nil
}

func gitTag() (string, error) {
	s, err := output("git", "describe", "--tags")
	if err != nil {
		ee, ok := errors.Cause(err).(*exec.ExitError)
		if ok && ee.Exited() {
			// probably no git tag
			return "dev", nil
		}
		return "", err
	}

	return strings.TrimSuffix(s, "\n"), nil
}

func shouldWork(ctx context.Context, env []string, dir string, cmdName string, args ...string) {
	loc, err := exec.LookPath(cmdName)
	qod.ANE(err)

	cmd := exec.CommandContext(ctx, loc, args...)
	cmd.Dir = dir
	cmd.Env = append(env, os.Environ()...)

	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr

	log.Printf("starting process, env: %v, pwd: %s, cmd: %s, args: %v", env, dir, loc, args)
	err = cmd.Run()
	qod.ANE(err)
}

func goBuild(ctx context.Context, env []string, dir string, pkgname string) {
	shouldWork(ctx, env, dir, "go", "build", "-v", pkgBase+pkgname)
}

func goInstall(ctx context.Context, env []string, pkgname string) {
	shouldWork(ctx, nil, wd, "go", "install", pkgBase+pkgname)
}

func goBuildPlugin(ctx context.Context, dir, pkgname, fname string) {
	if runtime.GOOS != "linux" {
		qod.Printlnf("plugins don't work on non-linux machines yet :(")
		return
	}

	shouldWork(ctx, nil, dir, "go", "build", "-v", "-buildmode=plugin", "-o="+fname, pkgBase+pkgname)
	qod.Printlnf("built %s for %s", fname, runtime.GOOS)
}

func buildBins(goos string) {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	d := filepath.Join(wd, "./bin")

	os.Mkdir(filepath.Join(d, goos), 0777)

	for _, pkg := range []string{"route-httpagent", "route-cli", "routed"} {
		goBuild(ctx, []string{"GOOS=" + goos}, filepath.Join(d, goos), "cmd/"+pkg)

		if goos == runtime.GOOS {
			goInstall(ctx, nil, "cmd/"+pkg)
		}

		qod.Printlnf("built binary for %s for os %s", pkg, goos)
	}
}

// Docker builds docker images
func Docker() {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	ver, err := gitTag()
	qod.ANE(err)

	shouldWork(ctx, nil, wd, "box", "box.rb")
	shouldWork(ctx, nil, wd, "docker", "tag", "xena/route:latest", "xena/route:"+ver)
	shouldWork(ctx, nil, wd, "docker", "tag", "xena/route:thick", "xena/route:thick-"+ver)
}

// Linux builds binaries for linux
func Linux() {
	buildBins("linux")
	Plugin()
}

// Windows builds binaries for windows
func Windows() {
	buildBins("windows")
}

// Darwin builds binaries for darwin
func Darwin() {
	buildBins("darwin")
}

// Build builds the binaries for route and routed.
func Build() {
	buildBins(runtime.GOOS)

	if runtime.GOOS == "linux" {
		Plugin()
	}
}

// Plugin builds all of the plugins for programs wanting to augment themselves with route.
func Plugin() {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	goBuildPlugin(ctx, filepath.Join(wd, "./bin/linux"), "plugins/autohttpagent", "autohttpagent.so")
}

// Package builds all binaries for all platforms and assembles a package of
// documentation, useful tools and all route components.
func Package() {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	os.RemoveAll("./bin/darwin")
	os.RemoveAll("./bin/linux")
	os.RemoveAll("./bin/windows")

	mg.Deps(Linux, Windows, Darwin)

	ver, err := gitTag()
	qod.ANE(err)

	for _, osname := range []string{"linux", "windows", "darwin"} {
		dir, err := ioutil.TempDir("", "route-package-"+osname)
		qod.ANE(err)
		qod.Printlnf("%s", dir)

		for _, file := range []string{"route-cli", "route-httpagent", "routed"} {
			if osname == "windows" {
				file = file + ".exe"
			}
			shouldWork(ctx, nil, filepath.Join(wd, "bin", osname), "cp", "-rf", file, dir)
		}

		shouldWork(ctx, nil, wd, "cp", "-rf", "doc", dir)
		shouldWork(ctx, nil, wd, "cp", "-rf", "README.md", dir)

		shouldWork(ctx, nil, dir, "tar", "czf", filepath.Join(wd, "bin", fmt.Sprintf("route-%s-%s.tar.gz", ver, osname)), ".")
	}
}

// Version is the version as git reports.
func Version() {
	ver, err := gitTag()
	qod.ANE(err)
	qod.Printlnf("route-%s", ver)
}

// Test runs all of the functional and unit tests for the project.
func Test() {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	shouldWork(ctx, nil, wd, "go", "test", "-v", "./...")
}
